import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Model
from tensorflow.keras.layers import *
from tensorflow.keras.mixed_precision import experimental as mixed_precision
# from tensorflow.keras.activations import relu


# def clipped_relu(x):
#     return relu(x, max_value=20)


def get_deepspeech2(input_dim=None, output_dim=29,
                    is_mixed_precision=True, random_state=1) -> keras.Model:
    """

    input_dim: int i wielokrotność 4
    output_dim: licba liter w słowniku

    """
    if is_mixed_precision:
        policy = mixed_precision.Policy('float32')
        mixed_precision.set_policy(policy)

    np.random.seed(random_state)
    tf.random.set_seed(random_state)

    # the input
    input_data = Input(name='the_input', shape=(None, input_dim), dtype='float32')

    # Batch normalize
    bn1 = BatchNormalization(axis=-1, name='BN_1')(input_data)

    # 1D Convs
    conv1 = Conv1D(512, 5, strides=1, activation='relu', name='Conv1D_1')(bn1)
    cbn1 = BatchNormalization(axis=-1, name='CBN_1')(conv1)
    conv2 = Conv1D(512, 5, strides=1, activation='relu', name='Conv1D_2')(cbn1)
    cbn2 = BatchNormalization(axis=-1, name='CBN_2')(conv2)
    conv3 = Conv1D(512, 5, strides=1, activation='relu', name='Conv1D_3')(cbn2)

    # Batch normalize
    x = BatchNormalization(axis=-1, name='BN_2')(conv3)

    # BiRNNs
    # birnn1 = Bidirectional(SimpleRNN(1280, return_sequences=True, name='BiRNN_1'), merge_mode='sum')(bn2)
    # birnn2 = Bidirectional(SimpleRNN(1280, return_sequences=True, name='BiRNN_2'), merge_mode='sum')(birnn1)
    # birnn3 = Bidirectional(SimpleRNN(1280, return_sequences=True, name='BiRNN_3'), merge_mode='sum')(birnn2)
    # birnn4 = Bidirectional(SimpleRNN(1280, return_sequences=True, name='BiRNN_4'), merge_mode='sum')(birnn3)
    # birnn5 = Bidirectional(SimpleRNN(1280, return_sequences=True, name='BiRNN_5'), merge_mode='sum')(birnn4)
    # birnn6 = Bidirectional(SimpleRNN(1280, return_sequences=True, name='BiRNN_6'), merge_mode='sum')(birnn5)
    # birnn7 = Bidirectional(SimpleRNN(1280, return_sequences=True, name='BiRNN_7'), merge_mode='sum')(birnn6)

    # BiRNNs
    for i in [1, 2, 3, 4, 5]:
        recurrent = GRU(units=800,
                        activation='tanh',
                        recurrent_activation='sigmoid',
                        use_bias=True,
                        return_sequences=True,
                        reset_after=True,
                        name=f'gru_{i}')
        x = Bidirectional(recurrent,
                          name=f'bidirectional_{i}',
                          merge_mode='concat')(x)
        x = Dropout(rate=0.5)(x) if i < 5 else x  # Only between

    # Batch normalize
    bn3 = BatchNormalization(axis=-1, name='BN_3')(x)

    dense = TimeDistributed(Dense(1024, activation='relu', name='FC1'))(bn3)
    y_pred = TimeDistributed(Dense(output_dim, activation='softmax', name='y_pred'))(dense)

    model = Model(inputs=input_data, outputs=y_pred)

    # # your ground truth data. The data you are going to compare with the model's outputs in training
    # labels = Input(name='the_labels', shape=[label_dim], dtype='float32')
    # # the length (in steps, or chars this case) of each sample (sentence) in the y_pred tensor
    # input_length = Input(name='input_length', shape=[1], dtype='float32')
    # #  the length (in steps, or chars this case) of each sample (sentence) in the y_true
    # label_length = Input(name='label_length', shape=[1], dtype='float32')
    # output = Lambda(ctc_loss, output_shape=(1,), name='ctc')([y_pred, labels, input_length, label_length])
    # model = Model(inputs=[input_data, labels, input_length, label_length], outputs=output, name="deepspeech2pro_v1")
    return model



# calculate filterbank features. Provides e.g. fbank and mfcc features for use in ASR applications
# Author: James Lyons 2012
from __future__ import division
import numpy
from . import sigproc
from scipy.fftpack import dct


def calculate_nfft(samplerate, winlen):
    """Calculates the FFT size as a power of two greater than or equal to
    the number of samples in a single window length.

    Having an FFT less than the window length loses precision by dropping
    many of the samples; a longer FFT than the window allows zero-padding
    of the FFT buffer which is neutral in terms of frequency domain conversion.

    :param samplerate: The sample rate of the signal we are working with, in Hz.
    :param winlen: The length of the analysis window in seconds.
    """
    window_length_samples = winlen * samplerate
    nfft = 1
    while nfft < window_length_samples:
        nfft *= 2
    return nfft


def mfcc(signal, samplerate=16000, winlen=0.025, winstep=0.01, numcep=13,
         nfilt=26, nfft=None, lowfreq=0, highfreq=None, preemph=0.97, ceplifter=22, appendEnergy=True,
         winfunc=lambda x: numpy.ones((x,))):
    """Compute MFCC features from an audio signal.

    :param signal: the audio signal from which to compute features. Should be an N*1 array
    :param samplerate: the sample rate of the signal we are working with, in Hz.
    :param winlen: the length of the analysis window in seconds. Default is 0.025s (25 milliseconds)
    :param winstep: the step between successive windows in seconds. Default is 0.01s (10 milliseconds)
    :param numcep: the number of cepstrum to return, default 13
    :param nfilt: the number of filters in the filterbank, default 26.
    :param nfft: the FFT size. Default is None, which uses the calculate_nfft function to choose the smallest size that does not drop sample data.
    :param lowfreq: lowest band edge of mel filters. In Hz, default is 0.
    :param highfreq: highest band edge of mel filters. In Hz, default is samplerate/2
    :param preemph: apply preemphasis filter with preemph as coefficient. 0 is no filter. Default is 0.97.
    :param ceplifter: apply a lifter to final cepstral coefficients. 0 is no lifter. Default is 22.
    :param appendEnergy: if this is true, the zeroth cepstral coefficient is replaced with the log of the total frame energy.
    :param winfunc: the analysis window to apply to each frame. By default no window is applied. You can use numpy window functions here e.g. winfunc=numpy.hamming
    :returns: A numpy array of size (NUMFRAMES by numcep) containing features. Each row holds 1 feature vector.
    """
    nfft = nfft or calculate_nfft(samplerate, winlen)
    feat, energy = fbank(signal, samplerate, winlen, winstep, nfilt, nfft, lowfreq, highfreq, preemph, winfunc)
    feat = numpy.log(feat)
    feat = dct(feat, type=2, axis=1, norm='ortho')[:, :numcep]
    feat = lifter(feat, ceplifter)
    if appendEnergy: feat[:, 0] = numpy.log(energy)  # replace first cepstral coefficient with log of frame energy
    return feat


def fbank(signal, samplerate=16000, winlen=0.025, winstep=0.01,
          nfilt=26, nfft=512, lowfreq=0, highfreq=None, preemph=0.97,
          winfunc=lambda x: numpy.ones((x,))):
    """Compute Mel-filterbank energy features from an audio signal.

    :param signal: the audio signal from which to compute features. Should be an N*1 array
    :param samplerate: the sample rate of the signal we are working with, in Hz.
    :param winlen: the length of the analysis window in seconds. Default is 0.025s (25 milliseconds)
    :param winstep: the step between successive windows in seconds. Default is 0.01s (10 milliseconds)
    :param nfilt: the number of filters in the filterbank, default 26.
    :param nfft: the FFT size. Default is 512.
    :param lowfreq: lowest band edge of mel filters. In Hz, default is 0.
    :param highfreq: highest band edge of mel filters. In Hz, default is samplerate/2
    :param preemph: apply preemphasis filter with preemph as coefficient. 0 is no filter. Default is 0.97.
    :param winfunc: the analysis window to apply to each frame. By default no window is applied. You can use numpy window functions here e.g. winfunc=numpy.hamming
    :returns: 2 values. The first is a numpy array of size (NUMFRAMES by nfilt) containing features. Each row holds 1 feature vector. The
        second return value is the energy in each frame (total energy, unwindowed)
    """
    highfreq = highfreq or samplerate / 2
    signal = sigproc.preemphasis(signal, preemph)
    frames = sigproc.framesig(signal, winlen * samplerate, winstep * samplerate, winfunc)
    pspec = sigproc.powspec(frames, nfft)
    energy = numpy.sum(pspec, 1)  # this stores the total energy in each frame
    energy = numpy.where(energy == 0, numpy.finfo(float).eps, energy)  # if energy is zero, we get problems with log

    fb = get_filterbanks(nfilt, nfft, samplerate, lowfreq, highfreq)
    feat = numpy.dot(pspec, fb.T)  # compute the filterbank energies
    feat = numpy.where(feat == 0, numpy.finfo(float).eps, feat)  # if feat is zero, we get problems with log

    return feat, energy


def logfbank(signal, samplerate=16000, winlen=0.025, winstep=0.01,
             nfilt=26, nfft=512, lowfreq=0, highfreq=None, preemph=0.97,
             winfunc=lambda x: numpy.ones((x,))):
    """Compute log Mel-filterbank energy features from an audio signal.

    :param signal: the audio signal from which to compute features. Should be an N*1 array
    :param samplerate: the sample rate of the signal we are working with, in Hz.
    :param winlen: the length of the analysis window in seconds. Default is 0.025s (25 milliseconds)
    :param winstep: the step between successive windows in seconds. Default is 0.01s (10 milliseconds)
    :param nfilt: the number of filters in the filterbank, default 26.
    :param nfft: the FFT size. Default is 512.
    :param lowfreq: lowest band edge of mel filters. In Hz, default is 0.
    :param highfreq: highest band edge of mel filters. In Hz, default is samplerate/2
    :param preemph: apply preemphasis filter with preemph as coefficient. 0 is no filter. Default is 0.97.
    :param winfunc: the analysis window to apply to each frame. By default no window is applied. You can use numpy window functions here e.g. winfunc=numpy.hamming
    :returns: A numpy array of size (NUMFRAMES by nfilt) containing features. Each row holds 1 feature vector.
    """
    feat, energy = fbank(signal, samplerate, winlen, winstep, nfilt, nfft, lowfreq, highfreq, preemph, winfunc)
    return numpy.log(feat)


def ssc(signal, samplerate=16000, winlen=0.025, winstep=0.01,
        nfilt=26, nfft=512, lowfreq=0, highfreq=None, preemph=0.97,
        winfunc=lambda x: numpy.ones((x,))):
    """Compute Spectral Subband Centroid features from an audio signal.

    :param signal: the audio signal from which to compute features. Should be an N*1 array
    :param samplerate: the sample rate of the signal we are working with, in Hz.
    :param winlen: the length of the analysis window in seconds. Default is 0.025s (25 milliseconds)
    :param winstep: the step between successive windows in seconds. Default is 0.01s (10 milliseconds)
    :param nfilt: the number of filters in the filterbank, default 26.
    :param nfft: the FFT size. Default is 512.
    :param lowfreq: lowest band edge of mel filters. In Hz, default is 0.
    :param highfreq: highest band edge of mel filters. In Hz, default is samplerate/2
    :param preemph: apply preemphasis filter with preemph as coefficient. 0 is no filter. Default is 0.97.
    :param winfunc: the analysis window to apply to each frame. By default no window is applied. You can use numpy window functions here e.g. winfunc=numpy.hamming
    :returns: A numpy array of size (NUMFRAMES by nfilt) containing features. Each row holds 1 feature vector.
    """
    highfreq = highfreq or samplerate / 2
    signal = sigproc.preemphasis(signal, preemph)
    frames = sigproc.framesig(signal, winlen * samplerate, winstep * samplerate, winfunc)
    pspec = sigproc.powspec(frames, nfft)
    pspec = numpy.where(pspec == 0, numpy.finfo(float).eps, pspec)  # if things are all zeros we get problems

    fb = get_filterbanks(nfilt, nfft, samplerate, lowfreq, highfreq)
    feat = numpy.dot(pspec, fb.T)  # compute the filterbank energies
    R = numpy.tile(numpy.linspace(1, samplerate / 2, numpy.size(pspec, 1)), (numpy.size(pspec, 0), 1))

    return numpy.dot(pspec * R, fb.T) / feat


def hz2mel(hz):
    """Convert a value in Hertz to Mels

    :param hz: a value in Hz. This can also be a numpy array, conversion proceeds element-wise.
    :returns: a value in Mels. If an array was passed in, an identical sized array is returned.
    """
    return 2595 * numpy.log10(1 + hz / 700.)


def mel2hz(mel):
    """Convert a value in Mels to Hertz

    :param mel: a value in Mels. This can also be a numpy array, conversion proceeds element-wise.
    :returns: a value in Hertz. If an array was passed in, an identical sized array is returned.
    """
    return 700 * (10 ** (mel / 2595.0) - 1)


def get_filterbanks(nfilt=20, nfft=512, samplerate=16000, lowfreq=0, highfreq=None):
    """Compute a Mel-filterbank. The filters are stored in the rows, the columns correspond
    to fft bins. The filters are returned as an array of size nfilt * (nfft/2 + 1)

    :param nfilt: the number of filters in the filterbank, default 20.
    :param nfft: the FFT size. Default is 512.
    :param samplerate: the sample rate of the signal we are working with, in Hz. Affects mel spacing.
    :param lowfreq: lowest band edge of mel filters, default 0 Hz
    :param highfreq: highest band edge of mel filters, default samplerate/2
    :returns: A numpy array of size nfilt * (nfft/2 + 1) containing filterbank. Each row holds 1 filter.
    """
    highfreq = highfreq or samplerate / 2
    assert highfreq <= samplerate / 2, "highfreq is greater than samplerate/2"

    # compute points evenly spaced in mels
    lowmel = hz2mel(lowfreq)
    highmel = hz2mel(highfreq)
    melpoints = numpy.linspace(lowmel, highmel, nfilt + 2)
    # our points are in Hz, but we use fft bins, so we have to convert
    #  from Hz to fft bin number
    bin = numpy.floor((nfft + 1) * mel2hz(melpoints) / samplerate)

    fbank = numpy.zeros([nfilt, nfft // 2 + 1])
    for j in range(0, nfilt):
        for i in range(int(bin[j]), int(bin[j + 1])):
            fbank[j, i] = (i - bin[j]) / (bin[j + 1] - bin[j])
        for i in range(int(bin[j + 1]), int(bin[j + 2])):
            fbank[j, i] = (bin[j + 2] - i) / (bin[j + 2] - bin[j + 1])
    return fbank


def lifter(cepstra, L=22):
    """Apply a cepstral lifter the the matrix of cepstra. This has the effect of increasing the
    magnitude of the high frequency DCT coeffs.

    :param cepstra: the matrix of mel-cepstra, will be numframes * numcep in size.
    :param L: the liftering coefficient to use. Default is 22. L <= 0 disables lifter.
    """
    if L > 0:
        nframes, ncoeff = numpy.shape(cepstra)
        n = numpy.arange(ncoeff)
        lift = 1 + (L / 2.) * numpy.sin(numpy.pi * n / L)
        return lift * cepstra
    else:
        # values of L <= 0, do nothing
        return cepstra


def delta(feat, N):
    """Compute delta features from a feature vector sequence.

    :param feat: A numpy array of size (NUMFRAMES by number of features) containing features. Each row holds 1 feature vector.
    :param N: For each frame, calculate delta features based on preceding and following N frames
    :returns: A numpy array of size (NUMFRAMES by number of features) containing delta features. Each row holds 1 delta feature vector.
    """
    if N < 1:
        raise ValueError('N must be an integer >= 1')
    NUMFRAMES = len(feat)
    denominator = 2 * sum([i ** 2 for i in range(1, N + 1)])
    delta_feat = numpy.empty_like(feat)
    padded = numpy.pad(feat, ((N, N), (0, 0)), mode='edge')  # padded version of feat
    for t in range(NUMFRAMES):
        delta_feat[t] = numpy.dot(numpy.arange(-N, N + 1),
                                  padded[t: t + 2 * N + 1]) / denominator  # [t : t+2*N+1] == [(N+t)-N : (N+t)+N+1]
    return delta_feat



import sys
from tinytag import TinyTag


# https://pypi.org/project/tinytag/

def get_file_tags(audio_file):
    tag = TinyTag.get(audio_file)
    print(tag.filesize, '|', tag.audio_offest, "|", tag.bitrate, "|", tag.channels, "|", tag.duration, "|",
          tag.samplerate, "|", tag.audio_offset)


if __name__ == "__main__":
    get_file_tags(sys.argv[1])



import numpy as np
from .mfcc import fbank
from . import feature_extractor


class FilterBanks(feature_extractor.FeaturesExtractor):

    def __init__(self, features_num: int, samplerate: int = 16000, is_standardization=True, **kwargs):
        self.features_num = features_num
        self.is_standardization = is_standardization
        self.params = kwargs
        self.samplerate = samplerate

    def make_features(self, audio: np.ndarray) -> np.ndarray:
        """ Use `python_speech_features` lib to extract log filter banks from
        the features file. """
        audio = self.normalize(audio.astype(np.float32))
        audio = (audio * np.iinfo(np.int16).max).astype(np.int16)
        feat, energy = fbank(
            audio, nfilt=self.features_num, samplerate=self.samplerate, **self.params
        )
        features = np.log(feat)
        return self.standardize(features) if self.is_standardization else features



import numpy as np
import pandas as pd
import tensorflow as tf
import deepasr as asr


def get_config(features, multi_gpu):
    alphabet_en = asr.vocab.Alphabet(lang='en')

    features_extractor = asr.features.preprocess(feature_type=features, features_num=161,
                                                 samplerate=16000,
                                                 winlen=0.02,
                                                 winstep=0.01,
                                                 winfunc=np.hanning)

    model = asr.model.get_deepspeech2(
        input_dim=161,
        output_dim=29,
        is_mixed_precision=True
    )
    optimizer = tf.keras.optimizers.Adam(
        lr=1e-4,
        beta_1=0.9,
        beta_2=0.999,
        epsilon=1e-8
    )
    decoder = asr.decoder.GreedyDecoder()

    pipeline = asr.pipeline.ctc_pipeline.CTCPipeline(
        alphabet=alphabet_en, features_extractor=features_extractor, model=model, optimizer=optimizer, decoder=decoder,
        sample_rate=16000, mono=True, multi_gpu=multi_gpu
    )
    return pipeline


def run(train_data, test_data, features='fbank', batch_size=32, epochs=10, multi_gpu=True):
    pipeline = get_config(features, multi_gpu)
    # history = pipeline.fit_iter(train_data, batch_size=batch_size, epochs=epochs, iter_num=1000)
    history = pipeline.fit_generator(train_data, batch_size=batch_size, epochs=epochs)
    pipeline.save('./checkpoints')
    print("Truth:", test_data['transcript'].to_list()[0])
    print("Prediction", pipeline.predict(test_data['path'].to_list()[0]))
    return history


def test_model(test_data):
    pipeline = asr.pipeline.load('checkpoints')
    print("Truth:", test_data['transcripts'].to_list()[0])
    print("Prediction", pipeline.predict(test_data['path'].to_list()[0]))


if __name__ == "__main__":
    train = pd.read_csv('train_data.csv')
    test = pd.read_csv('test_data.csv')
    run(train, test, features='fbank', batch_size=32, epochs=100, multi_gpu=False)
    # test_model(test)



from typing import List, Iterable, Tuple, Union
from collections import namedtuple
import pandas as pd
from . import distance
# from .. import dataset
from .. import pipeline

Metric = namedtuple('Metric', ['transcript', 'prediction', 'wer', 'cer'])


def calculate_error_rates(ctc_pipeline: pipeline.Pipeline,
                          data: pd.DataFrame,
                          return_metrics: bool = False
                          ) -> Union[Tuple[float, float], pd.DataFrame]:
    """ Calculate base metrics: WER and CER. """
    metrics = []
    for audio, transcript in zip(data['path'].values, data['transcripts'].values):
        prediction = ctc_pipeline.predict(audio)
        batch_metrics = get_metrics(sources=prediction,
                                    destinations=[transcript])
        metrics.extend(batch_metrics)
    metrics = pd.DataFrame(metrics)
    return metrics if return_metrics else (metrics.wer.mean(), metrics.cer.mean())


def get_metrics(sources: List[str],
                destinations: List[str]) -> Iterable[Metric]:
    """ Calculate base metrics in one batch: WER and CER. """
    for source, destination in zip(sources, destinations):
        wer_distance, *_ = distance.edit_distance(source.split(),
                                                  destination.split())
        wer = wer_distance / len(destination.split())

        cer_distance, *_ = distance.edit_distance(list(source),
                                                  list(destination))
        cer = cer_distance / len(destination)
        yield Metric(destination, source, wer, cer)


def get_cer(source: str, destination: str) -> float:
    cer_distance, *_ = distance.edit_distance(list(source),
                                              list(destination))
    cer = cer_distance / len(destination)

    return cer


def get_wer(source: str, destination: str) -> float:
    wer_distance, *_ = distance.edit_distance(source.split(),
                                              destination.split())
    wer = wer_distance / len(destination.split())

    return wer



import functools
import operator
from typing import Callable, List, Union, Tuple
import h5py
import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow import keras
from . import evaluate
# from .. import dataset
from .. import pipeline
from .. import utils


def save_metrics_and_activations(pipeline: pipeline.Pipeline,
                                 data: pd.DataFrame,
                                 store_path: str,
                                 prepared_features: bool = False,
                                 return_metrics: bool = False
                                 ) -> Union[Tuple[float, float], pd.DataFrame]:
    columns = ['sample_id', 'transcript', 'prediction', 'wer', 'cer']
    references = pd.DataFrame(columns=columns).set_index('sample_id')
    get_activations = get_activations_function(pipeline.model)

    with h5py.File(store_path, mode='w') as store:
        for audio, transcript in zip(data['path'].values, data['transcripts'].values):
            features = audio if prepared_features else pipeline.features_extractor([utils.read_audio(audio)])
            *activations, y_hat = get_activations([features, 0])
            decoded_labels = pipeline.decoder(y_hat)
            predictions = pipeline.alphabet.get_batch_transcripts(decoded_labels)
            batch_metrics = list(evaluate.get_metrics(sources=predictions,
                                                      destinations=transcript))

            save_in_store(store, [*activations, y_hat], batch_metrics, references)

    with pd.HDFStore(store_path, mode='r+') as store:
        store.put('references', references)
    metrics = pd.DataFrame(functools.reduce(operator.concat, batch_metrics))
    return metrics if return_metrics else (metrics.wer.mean(), metrics.cer.mean())


def get_activations_function(model: keras.Model) -> Callable:
    """ Function which handle all activations through one pass. """
    inputs = [model.input, tf.keras.learning_phase()]
    outputs = [layer.output for layer in model.layers][1:]
    return tf.keras.function(inputs, outputs)


def save_in_store(store: h5py.File,
                  layer_outputs: List[np.ndarray],
                  metrics: List[evaluate.Metric],
                  references: pd.DataFrame):
    """ Save batch data into HDF5 file. """
    for index, metric in enumerate(metrics):
        sample_id = len(references)
        references.loc[sample_id] = metric
        for output_index, batch_layer_outputs in enumerate(layer_outputs):
            layer_output = batch_layer_outputs[index]
            store.create_dataset(f'outputs/{output_index}/{sample_id}', data=layer_output)



from .filter_banks import FilterBanks
from .spectrogram import Spectrogram


def preprocess(feature_type: str = 'fbank', features_num: int = 161,
               samplerate: int = 16000,
               winlen: float = 0.02,
               winstep: float = 0.01,
               winfunc=None,
               is_standardization=True,
               pad_audio_to: int = 0):
    ''' This method extracts the audio features based on fbank or spectrogram '''
    if feature_type == 'fbank':
        features_extractor = FilterBanks(features_num=features_num, samplerate=samplerate, winlen=winlen,
                                         winstep=winstep, winfunc=winfunc,
                                         is_standardization=is_standardization)
    else:
        features_extractor = Spectrogram(
            features_num=features_num,
            samplerate=samplerate,
            winlen=winlen,
            winstep=winstep,
            winfunc=winfunc,
            pad_audio_to=pad_audio_to
        )

    return features_extractor
